//#define _DEBUG

using System;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Diagnostics;
using System.Collections;
using System.Text;
using System.Text.RegularExpressions;

namespace Mail.Providers
{
	/// <summary>
	///   POP3
	/// </summary>	
	/// <example>  .
	/// <code>
	/// using (Pop3 pop3 = new Pop3("host"))
	///	{
	///		pop3.LogIn("Username", "Password");
	///		Console.WriteLine(" :" + pop3.NumberOfMessages);
	///		
	///		using(Message msg = pop3.GetMessage(1)) //   
	///		{
	///			Console.WriteLine(": " + msg.Subject); 
	///		}
	///	}
	/// </code>
	/// </example>	
	public class Pop3 : Provider
	{
		#region Constants
		const int MaxReceiveSize = 1024;
		const int POP3DefaultPort = 110;

		const int SendTimeout = 60000; //  .
        public int ReceiveTimeout = 2000000; //  .
        public int PollTimeout = 100000; //  .

		const string CRLF = "\r\n";
		
		const string FiveOctalTerm = "\r\n.\r\n";
		const string STAT_OK = "+OK";
		const string STAT_ERR = "-ERR";

		const char SPACE = ' ';
		const char CR = '\r';
		#endregion

		Socket _socket = null;

        MaildropStatus _status = null;
		uint [] _messagelist = null;

		bool _authenticated = false;

		#region DEBUG functions
		FileStream GetDebugStream()
		{
			return new FileStream(@"C:\trace_response.log", FileMode.Append);
		}
		[Conditional("_DEBUG")]
		void Trace(byte [] str)
		{
			FileStream fs = GetDebugStream();
			fs.Write(str, 0, str.Length);
			fs.Close();
		}

		[Conditional("_DEBUG")]
		void Trace(string str)
		{
			FileStream fs = GetDebugStream();
			fs.Write(Encoding.ASCII.GetBytes(str), 0, str.Length);
			fs.Close();
		}
		#endregion

		#region Constructors 
		/// <summary>
        ///       (110)       
        ///     .		
		/// </summary>
		/// <param name="server">IP  .</param>
		public Pop3(string server)
		{
			_server = server;
			_port = POP3DefaultPort;
			TempDirectory = Path.GetTempPath();
		}

		/// <summary>
        ///        (110).
		/// </summary>
        /// <param name="server">IP  .</param>
		/// <param name="temp">  .</param>
		public Pop3(string server, string temp)
		{
			_server = server;
			_port = POP3DefaultPort;
			TempDirectory = temp;
		}

		/// <summary>
		///   .
		/// </summary>
        /// <param name="server">IP  .</param>
		/// <param name="port"> .</param>
		public Pop3(string server, int port)
		{
			_server = server;
			_port = port;
			TempDirectory = Path.GetTempPath();
		}

		/// <summary>
        ///   .
		/// </summary>
        /// <param name="server">IP  .</param>
        /// <param name="port"> .</param>
        /// <param name="temp">  .</param>
		public Pop3(string server, int port, string temp)
		{
			_server = server;
			_port = port;
			TempDirectory = temp;
		}
		#endregion

		#region Public properties
		/// <summary>
		///    true,      .
		/// </summary>
		public bool IsMessages
		{
			get 
			{
				return (NumberOfMessages > 0);
			}
		}

		/// <summary>
		///    .
		/// </summary>
		public uint NumberOfMessages
		{
			get 
			{
				// not initialized
				if (_status == null)
				{
					GetStatus();
				}

				return _status.messages;
			}	
		}
		#endregion

		#region Method-Property substitution
        /// <summary>
        ///   .
        /// </summary>
        /// <returns></returns>
		public uint GetNumberOfMessages()
		{
            return NumberOfMessages;
		}
        
		public bool GetIsMessages()
		{
			return IsMessages;
		}
		#endregion

		/// <summary>
        ///    ,        STAT.		
		/// </summary>
        /// <example> STAT.          "+OK", 
        ///              . 
        /// ,         . 
        /// </example>
		void GetStatus()
		{
			CheckConnection();

			Send("STAT");
			string tmp = Receive();

			string [] tokens = tmp.Split(new Char[] {SPACE, CR}, 4);

			try
			{
				_status = new MaildropStatus(
					Convert.ToUInt32(tokens[1], 10), 
					Convert.ToUInt32(tokens[2], 10)
				);
			} 
			catch (Exception e)
			{
                throw new CoreException("  ", e);
			}
		}
        /// <summary>
        ///    .
        /// </summary>
        /// <param name="server"> .</param>
        /// <param name="port"> .</param>
		void EstablishConnection(string server, int port)
		{
			//  IP  .
			IPAddress ipadr = Dns.Resolve(server).AddressList[0];
			IPEndPoint ephost = new IPEndPoint(ipadr, port);
 
			//  Socket      TCP.
			_socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);

			
			LingerOption linger = new LingerOption(true, 10);
			_socket.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.Linger, linger);
			//   .
			_socket.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.SendTimeout, SendTimeout);
			_socket.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.ReceiveTimeout, ReceiveTimeout);

			//   .
			_socket.Connect(ephost);
			if (!_socket.Connected)
			{
				throw new CoreException("  : " + server);
			}
		}

		/// <summary>
		///     .
		/// </summary>
		void CheckConnection()
		{
			if (_socket == null || !_socket.Connected) 
			{
				throw new CoreException("  .");
			}

			if (!_authenticated)
			{
                throw new CoreException("   ( LogIn). ");
			}
		}
        /// <summary>
        ///    .
        /// </summary>
        /// <param name="command"> .</param>
		void Send(string command)
		{
            //     CRLF.
			command += CRLF;
            //     ASCII.
			Encoding tmp = Encoding.ASCII;
			Byte [] buf = tmp.GetBytes(command);

			int total = buf.Length;
			while (total > 0)
			{
				total -= _socket.Send(buf, command.Length, SocketFlags.None);
			}
		}
        /// <summary>
        ///  POP3 .
        /// </summary>
        /// <param name="str">.</param>
		void AnalyseResponse(string str)
		{
			Trace(str);
//			Debug.WriteLine(str);

			if (str.StartsWith(STAT_ERR))
			{
				string msg;				
				int i = str.IndexOf(CRLF);
				if (i < 0) 
				{
					msg = " : " + STAT_ERR;
				} 
				else
				{
					//    ,   .
					msg = str.Substring(STAT_ERR.Length + 1, Math.Min(i - STAT_ERR.Length - 1, 79));
				}

				throw new ResponseException(msg);
			}
		}

		/// <summary>
		///    POP3 .
		/// </summary>
		/// <param name="index"> .</param>
		/// <returns></returns>
		public StringReader DumpMessage(int index)
		{
			CheckConnection();

			Send("RETR " + index);
			return new StringReader(Receive());
		}
        /// <summary>
        ///    .
        /// </summary>
        /// <param name="message">.</param>
        /// <returns></returns>
		string TruncateTail(string message)
		{
			if (!message.EndsWith(FiveOctalTerm))
			{
				Debug.WriteLine(" 5 : {" + message.Substring(message.Length - 5) + "}");
				throw new ResponseException("   .");
			}

            return message.Remove(message.Length - FiveOctalTerm.Length, FiveOctalTerm.Length);
		}

		/// <summary>
		///    .
		/// </summary>
		/// <returns>    .</returns>
        /// <example> LIST.      ,    . 
        /// ,     . 
        /// </example>
		public uint[] ListMessages()
		{
			CheckConnection();

			if (_messagelist == null)
			{
				Send("LIST");
				string tmp = Receive();
				tmp = TruncateTail(tmp);
				int start = tmp.IndexOf(CRLF);
				if (start > 0)
				{
					start += CRLF.Length;
					ArrayList l = new ArrayList();
					Regex r = new Regex(@"\r\n");
					string [] list = r.Split(tmp.Substring(start));
					if (list.Length > 0)
					{
						foreach (string s in list)
						{
							string [] f = s.Split(new char [] {' '}, 2);
							l.Add(Convert.ToUInt32(f[0], 10));
						}
					}

					if (l.Count > 0)
					{
						_messagelist = (uint [])l.ToArray(typeof(uint));
					}
					else
						_messagelist = new uint[0];
				} else 
					_messagelist = new uint[0];
			}

			return _messagelist;
		}

		/// <summary>
		///   NOOP  . 
		/// </summary>
		/// <remarks>
		/// <para>     .</para>
		/// </remarks>
        /// <example> NOOP. POP3        . 
        /// </example>
		public void SendNoop()
		{
			CheckConnection();

			Send("NOOP");
			string tmp = Receive();
		}

		/// <summary>
		///    .
		/// </summary>
		/// <remarks>
        /// <para>
        ///     ,   .
		/// </para>
		/// </remarks>
		/// <param name="index"> .</param>
		/// <returns>  .</returns>
		public string GetMessageUniqueID(uint index)
		{
			CheckConnection();

			Send("UIDL " + index);
			string tmp = Receive();

			string [] f = tmp.Split(new char [] {' ', '\r', '\n'}, 4);
            return f[2];            
		}

		/// <summary>
		///   .
		/// </summary>		
		/// <param name="index">.</param>
		/// <param name="liens">  .</param>
		/// <returns>   .</returns>
        /// <example> TOP.    , 
        ///            .
        /// </example>
		public Message GetMessageHeader(uint index, int top)
		{
			CheckConnection();

			Send("TOP " + index + " " + top);
			string message = Receive();

			message = Utils.RemoveByteStuffedSequence(message);
			return new Message(this, TruncateTail(message), index);
		}

		/// <summary> .
		/// </summary>
		/// <param name="index"> .</param>
        /// <example> DELETE. POP3      , 
        ///    ,       UPDATE. 
        /// </example>
		public override void DeleteMessage(uint index)
		{
			CheckConnection();

			Send("DELE " + index);
			string tmp = Receive();
		}

		/// <summary>
		///  .
		/// </summary>
		/// <param name="index"> .</param>
		/// <returns>.</returns>
        /// <example> RETR.       .        
        /// </example>
		public override Message GetMessage(uint index)
		{
			CheckConnection();

            Send("RETR " + index);
            string message = ReceiveMessage();

			message = Utils.RemoveByteStuffedSequence(message);
            return new Message(this, TruncateTail(message), index);
		}

		public void OnRecievedData( IAsyncResult ar )
		{
		}

        /// <summary>
        ///      .
        /// </summary>
        /// <returns> .</returns>
		StringBuilder UnsafeReceive()
		{
			StringBuilder tmp = new StringBuilder();
			Encoding cenc = Encoding.ASCII;
			IAsyncResult asynResult;
			byte[] buf = new byte[1024];
			int recv = 0;			
			do
			{
				asynResult = _socket.BeginReceive(buf, 0, buf.Length, SocketFlags.None, null, null);
				if (asynResult.AsyncWaitHandle.WaitOne())
				{
					recv = _socket.EndReceive(asynResult);
					string t = cenc.GetString(buf, 0, recv);
					tmp.Append(t);
					if (t.LastIndexOf(FiveOctalTerm) > 0) break;
				}
			}
			while(_socket.Poll(PollTimeout, SelectMode.SelectRead));

            return tmp;
		}
		/// <summary>
		///      .
		/// </summary>
		/// <returns> .</returns>
		StringBuilder UnsafeReceiveMessage()
		{
			StringBuilder tmp = new StringBuilder();
			Encoding cenc = Encoding.ASCII;
			IAsyncResult asynResult;
			byte[] buf = new byte[1024];
			int recv = 0;			
			do
			{
				asynResult = _socket.BeginReceive(buf, 0, buf.Length, SocketFlags.None, null, null);
				if (asynResult.AsyncWaitHandle.WaitOne())
				{
					recv = _socket.EndReceive(asynResult);
					string t = cenc.GetString(buf, 0, recv);
					tmp.Append(t);
					//if (t.LastIndexOf(FiveOctalTerm) > 0) 
					//	break;
				}
			}
			while(!tmp.ToString().EndsWith(FiveOctalTerm));

			return tmp;
		}
		/// <summary>
		///    .
		/// </summary>
		/// <returns> .</returns>
		string Receive()
		{
			StringBuilder tmp = UnsafeReceive();
            string str = tmp.ToString();
			AnalyseResponse(str);

			return str;
		}
		/// <summary>
		///     .
		/// </summary>
		/// <returns></returns>
		string ReceiveMessage()
		{
			StringBuilder tmp = UnsafeReceiveMessage();
			string str = tmp.ToString();
			AnalyseResponse(str);

			return str;
		}
        /// <summary>
        ///  .
        /// </summary>
        /// <param name="username"> .</param>
        /// <param name="password">.</param>
        /// <example>        .
        ///      ,   USER  PASS. 
        ///     USER,        . 
        ///    ,      PASS,    .
        /// <code>
        /// Client: USER username
        /// Server: +OK username
        /// Client: PASS mypass
        /// Server: +OK username
        /// </code>
        /// </example>
		void AuthenticateYourSelf(string username, string password)
		{
			Send("USER " + username);
			Receive();
            Send("PASS " + password);
			Receive();

			_authenticated = true;
		}

		/// <summary>
        ///      .
		/// </summary>
		/// <param name="username"> .</param>
		/// <param name="password">.</param>
		public override void LogIn(string username, string password)
		{
			try
			{
				if (_socket != null) 
				{
					Quit();
					ResetVariables();
				}
                //  .
				EstablishConnection(_server, _port);
				Receive(); //    .
				AuthenticateYourSelf(username, password);
			}
			catch (ShutdownException e)
			{
                throw new CoreException("   .", e);
			}
			catch (Exception e)
			{
				throw new CoreException(" ", e);
			}
		}

		/// <summary>
		///    .
		/// </summary>
		public override void Quit()
		{
			try
			{
				CheckConnection();
                //   POP3      UPDATE.
				Send("QUIT"); //    
			}
			catch (Exception e)
			{
				throw new ShutdownException("  ", e);
			}

			CloseSocket();
		}

		/// <summary>
		///   .
		/// </summary>
		void CloseSocket()
		{
			try
			{
				_socket.Shutdown(SocketShutdown.Both);
				_socket.Close();
				//  'Connected'    false   .
				if (_socket.Connected) 
				{
					throw new CoreException("  socket  : " + 
						Convert.ToString(System.Runtime.InteropServices.Marshal.GetLastWin32Error()));
				}

				_socket = null;
			}
			catch (SocketException e)
			{
                throw new CoreException("  socket", e);
			}
		}
        /// <summary>
        ///  .
        /// </summary>
		void ResetVariables()
		{
			_authenticated = false;
			_status = null;
			_messagelist = null;
		}

		/// <summary>
		///  .
		/// </summary>
		public override void Dispose()
		{
			try
			{
				Quit();
				ResetVariables();
			} 
			catch //    .
			{
				Debug.WriteLine("  socket");
			}

			GC.SuppressFinalize(this);
		}
	}
}
